defmodule Todo.Server do
  use GenServer

  # Expiry constants
  @expiry_timeout :timer.seconds(5)

  def start_link(name) do
    IO.puts "Starting to-do server for #{name}"
    GenServer.start_link(Todo.Server, name, name: via_tuple(name))
  end

  def add_entry(todo_server, new_entry) do
    GenServer.cast(todo_server, {:add_entry, new_entry})
  end

  def entries(todo_server, date) do
    GenServer.call(todo_server, {:entries, date})
  end

  def update_entry(todo_server, entry_id, updater_fun) do
    GenServer.cast(todo_server, {:update_entry, entry_id, updater_fun})
  end

  def delete_entry(todo_server, entry_id) do
    GenServer.cast(todo_server, {:delete_entry, entry_id})
  end


  def whereis(name) do
    :gproc.whereis_name({:n, :l, {:todo_server, name}})
  end

  defp via_tuple(name) do
    {:via, :gproc, {:n, :l, {:todo_server, name}}}
  end

  def init(name) do
    {:ok, {name, Todo.List.new}, @expiry_timeout}
  end


  def handle_cast({:add_entry, new_entry}, {name, todo_list}) do
    new_list = Todo.List.add_entry(todo_list, new_entry)
    Todo.Database.store({name, new_entry.date}, Todo.List.entries(new_list, new_entry.date))
    {:noreply, {name, new_list}, @expiry_timeout}
  end

  def handle_cast({:update_entry, entry_id, updater_fun}, {name, todo_list}) do
    {:noreply, {name, Todo.List.update_entry(todo_list, entry_id, updater_fun)}}
  end

  def handle_cast({:delete_entry, entry_id}, {name, todo_list}) do
    {:noreply, {name, Todo.List.delete_entry(todo_list, entry_id)}}
  end


  def handle_call({:entries, date}, _, {name, todo_list}) do
    {entries, todo_list} = case Todo.List.entries(todo_list, date) do
      nil ->
        entries = Todo.Database.get({name, date}) || []
        {
          entries,
          Todo.List.set_entries(todo_list, date, entries)
        }
      found -> {found, todo_list}
    end

    {:reply, entries, {name, todo_list}, @expiry_timeout}
  end

  # Handles timeout
  def handle_info(:timeout, {name, _} = state) do
    IO.puts "To-do server '#{name}' expired"
    {:stop, :normal, state}
  end

  def handle_info(_, state) do
    {:noreply, state, @expiry_timeout}
  end
end